@@ -1,4 +1,5 @@
 use std::collections::HashMap;
+use std::hash::{DefaultHasher, Hash, Hasher};
 use std::io::Read;
 use std::path::{Path, PathBuf};
 use std::sync::atomic::{AtomicBool, Ordering};
@@ -8,7 +9,7 @@ use cargo::core::compiler::{unit_graph::UnitDep, unit_graph::UnitGraph, Executor
 use cargo::core::profiles::Profiles;
 use cargo::core::{FeatureValue, Package, PackageId, Target, TargetKind, Workspace};
 use cargo::ops::{self, CompileFilter, CompileOptions, FilterRule, LibRule};
-use cargo::util::command_prelude::{ArgMatches, ArgMatchesExt, CompileMode, ProfileChecking};
+use cargo::util::command_prelude::{ArgMatches, ArgMatchesExt, ProfileChecking, UserIntent};
 use cargo::util::interning::InternedString;
 use cargo::{CliResult, GlobalContext};
 
@@ -16,7 +17,6 @@ use anyhow::Context as _;
 use cargo_util::paths::{copy, create_dir_all, open, read, read_bytes, write};
 use implib::def::ModuleDef;
 use implib::{Flavor, ImportLibrary, MachineType};
-use itertools::Itertools;
 use semver::Version;
 
 use crate::build_targets::BuildTargets;
@@ -28,7 +28,7 @@ use crate::target;
 fn build_include_file(
     ws: &Workspace,
     name: &str,
-    version: &Version,
+    version: Option<&Version>,
     root_output: &Path,
     root_path: &Path,
 ) -> anyhow::Result<()> {
@@ -42,15 +42,17 @@ fn build_include_file(
 
     // TODO: map the errors
     let mut config = cbindgen::Config::from_root_or_default(crate_path);
-    let warning = config.autogen_warning.unwrap_or_default();
-    let version_info = format!(
-        "\n#define {0}_MAJOR {1}\n#define {0}_MINOR {2}\n#define {0}_PATCH {3}\n",
-        name.to_uppercase().replace('-', "_"),
-        version.major,
-        version.minor,
-        version.patch
-    );
-    config.autogen_warning = Some(warning + &version_info);
+    if let Some(version) = version {
+        let warning = config.autogen_warning.unwrap_or_default();
+        let version_info = format!(
+            "\n#define {0}_MAJOR {1}\n#define {0}_MINOR {2}\n#define {0}_PATCH {3}\n",
+            name.to_uppercase().replace('-', "_"),
+            version.major,
+            version.minor,
+            version.patch
+        );
+        config.autogen_warning = Some(warning + &version_info);
+    }
     cbindgen::Builder::new()
         .with_crate(crate_path)
         .with_config(config)
@@ -111,7 +113,11 @@ fn patch_target(
     let manifest = pkg.manifest_mut();
     let targets = manifest.targets_mut();
 
-    let mut kinds = Vec::with_capacity(2);
+    let mut kinds = Vec::with_capacity(3);
+
+    if library_types.rlib {
+        kinds.push(CrateType::Rlib);
+    }
 
     if library_types.staticlib {
         kinds.push(CrateType::Staticlib);
@@ -235,6 +241,7 @@ struct FingerPrint {
     build_targets: BuildTargets,
     install_paths: InstallPaths,
     static_libs: String,
+    hasher: DefaultHasher,
 }
 
 #[derive(serde::Serialize, serde::Deserialize)]
@@ -249,21 +256,24 @@ impl FingerPrint {
         root_output: &Path,
         build_targets: &BuildTargets,
         install_paths: &InstallPaths,
+        capi_config: &CApiConfig,
     ) -> Self {
+        let mut hasher = DefaultHasher::new();
+
+        capi_config.hash(&mut hasher);
+
         Self {
             id: id.to_owned(),
             root_output: root_output.to_owned(),
             build_targets: build_targets.clone(),
             install_paths: install_paths.clone(),
             static_libs: String::new(),
+            hasher,
         }
     }
 
     fn hash(&self) -> anyhow::Result<Option<String>> {
-        use std::collections::hash_map::DefaultHasher;
-        use std::hash::{Hash, Hasher};
-
-        let mut hasher = DefaultHasher::new();
+        let mut hasher = self.hasher.clone();
         self.install_paths.hash(&mut hasher);
 
         let mut paths: Vec<&PathBuf> = Vec::new();
@@ -324,7 +334,7 @@ impl FingerPrint {
     }
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct CApiConfig {
     pub header: HeaderCApiConfig,
     pub pkg_config: PkgConfigCApiConfig,
@@ -332,15 +342,16 @@ pub struct CApiConfig {
     pub install: InstallCApiConfig,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct HeaderCApiConfig {
     pub name: String,
     pub subdirectory: String,
     pub generation: bool,
     pub enabled: bool,
+    pub emit_version_constants: bool,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct PkgConfigCApiConfig {
     pub name: String,
     pub filename: String,
@@ -351,14 +362,14 @@ pub struct PkgConfigCApiConfig {
     pub strip_include_path_components: usize,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub enum VersionSuffix {
     Major,
     MajorMinor,
     MajorMinorPatch,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct LibraryCApiConfig {
     pub name: String,
     pub version: Version,
@@ -388,19 +399,19 @@ impl LibraryCApiConfig {
     }
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, Hash)]
 pub struct InstallCApiConfig {
     pub include: Vec<InstallTarget>,
     pub data: Vec<InstallTarget>,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub enum InstallTarget {
     Asset(InstallTargetPaths),
     Generated(InstallTargetPaths),
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, Hash)]
 pub struct InstallTargetPaths {
     /// pattern to feed to glob::glob()
     ///
@@ -535,6 +546,11 @@ fn load_manifest_capi_config(
                 .and_then(|h| h.get("enabled"))
                 .map(|v| v.clone().try_into())
                 .unwrap_or(Ok(true))?,
+            emit_version_constants: header
+                .as_ref()
+                .and_then(|h| h.get("emit_version_constants"))
+                .map(|v| v.clone().try_into())
+                .unwrap_or(Ok(true))?,
         }
     } else {
         HeaderCApiConfig {
@@ -542,6 +558,7 @@ fn load_manifest_capi_config(
             subdirectory: String::from(name),
             generation: true,
             enabled: true,
+            emit_version_constants: true,
         }
     };
 
@@ -757,11 +774,11 @@ fn compile_options(
     gctx: &GlobalContext,
     args: &ArgMatches,
     profile: InternedString,
-    compile_mode: CompileMode,
+    compile_intent: UserIntent,
 ) -> anyhow::Result<CompileOptions> {
     use cargo::core::compiler::CompileKind;
     let mut compile_opts =
-        args.compile_options(gctx, compile_mode, Some(ws), ProfileChecking::Custom)?;
+        args.compile_options(gctx, compile_intent, Some(ws), ProfileChecking::Custom)?;
 
     compile_opts.build_config.requested_profile = profile;
 
@@ -843,7 +860,7 @@ impl Executor for Exec {
     }
 }
 
-use cargo::core::compiler::{unit_graph, UnitInterner};
+use cargo::core::compiler::{unit_graph, CompileMode, UnitInterner};
 use cargo::ops::create_bcx;
 
 fn set_deps_args(
@@ -980,7 +997,13 @@ impl CPackage {
             args.get_flag("meson"),
         )?;
 
-        let finger_print = FingerPrint::new(&id, root_output, &build_targets, &install_paths);
+        let finger_print = FingerPrint::new(
+            &id,
+            root_output,
+            &build_targets,
+            &install_paths,
+            &capi_config,
+        );
 
         Ok(CPackage {
             version,
@@ -1008,6 +1031,7 @@ fn deprecation_warnings(ws: &Workspace, args: &ArgMatches) -> anyhow::Result<()>
 pub struct LibraryTypes {
     pub staticlib: bool,
     pub cdylib: bool,
+    pub rlib: bool,
 }
 
 impl LibraryTypes {
@@ -1022,18 +1046,22 @@ impl LibraryTypes {
         // - https://github.com/lu-zero/cargo-c/issues/180
         Self {
             staticlib: true,
-            cdylib: target.os != "none" && target.env != "musl",
+            cdylib: target.os != "none",
+            rlib: false,
         }
     }
 
-    fn from_args(target: &target::Target, args: &ArgMatches) -> Self {
-        match args.get_many::<String>("library-type") {
-            Some(library_types) => Self::from_library_types(target, library_types),
-            None => Self::from_target(target),
+    fn from_args(target: &target::Target, args: &ArgMatches, want_rlib: bool) -> Self {
+        Self {
+            rlib: want_rlib,
+            ..match args.get_many::<String>("library-type") {
+                Some(library_types) => Self::from_library_types(target, library_types),
+                None => Self::from_target(target),
+            }
         }
     }
 
-    pub(crate) fn from_library_types<S: AsRef<str>>(
+    fn from_library_types<S: AsRef<str>>(
         target: &target::Target,
         library_types: impl Iterator<Item = S>,
     ) -> Self {
@@ -1048,7 +1076,11 @@ impl LibraryTypes {
         // forcing a cdylib for musl is allowed here (see [`LibraryTypes::from_target`])
         cdylib &= target.os != "none";
 
-        Self { staticlib, cdylib }
+        Self {
+            staticlib,
+            cdylib,
+            rlib: false,
+        }
     }
 
     const fn only_staticlib(self) -> bool {
@@ -1070,7 +1102,6 @@ fn static_libraries(link_line: &str, rustc_target: &target::Target) -> String {
             }
             !s.is_empty()
         })
-        .unique()
         .map(|lib| {
             if rustc_target.env == "msvc" && lib.ends_with(".lib") {
                 return format!("-l{}", lib.trim_end_matches(".lib"));
@@ -1087,6 +1118,7 @@ pub fn cbuild(
     config: &GlobalContext,
     args: &ArgMatches,
     default_profile: &str,
+    tests: bool,
 ) -> anyhow::Result<(Vec<CPackage>, CompileOptions)> {
     deprecation_warnings(ws, args)?;
 
@@ -1098,13 +1130,13 @@ pub fn cbuild(
 
     let rustc_target = target::Target::new(Some(&target), is_target_overridden)?;
 
-    let library_types = LibraryTypes::from_args(&rustc_target, args);
+    let library_types = LibraryTypes::from_args(&rustc_target, args, tests);
 
     let profile = args.get_profile_name(default_profile, ProfileChecking::Custom)?;
 
     let profiles = Profiles::new(ws, profile)?;
 
-    let mut compile_opts = compile_options(ws, config, args, profile, CompileMode::Build)?;
+    let mut compile_opts = compile_options(ws, config, args, profile, UserIntent::Build)?;
 
     // TODO: there must be a simpler way to get the right path.
     let root_output = ws
@@ -1134,7 +1166,7 @@ pub fn cbuild(
     for m in ws.members_mut().filter(|p| is_relevant_package(p)) {
         let cpkg = CPackage::from_package(m, args, library_types, &rustc_target, &root_output)?;
 
-        pristine |= cpkg.finger_print.load_previous().is_err();
+        pristine |= cpkg.finger_print.load_previous().is_err() || !cpkg.finger_print.is_valid();
 
         members.push(cpkg);
     }
@@ -1177,8 +1209,7 @@ pub fn cbuild(
 
     for cpkg in members.iter_mut() {
         // it is a new build, build the additional files and update update the cache
-        // if the hash value does not match.
-        if new_build && !cpkg.finger_print.is_valid() {
+        if new_build {
             let name = &cpkg.capi_config.library.name;
             let (pkg_config_static_libs, static_libs) = if library_types.only_cdylib() {
                 (String::new(), String::new())
@@ -1206,11 +1237,12 @@ pub fn cbuild(
 
             if capi_config.header.enabled {
                 let header_name = &capi_config.header.name;
+                let emit_version_constants = capi_config.header.emit_version_constants;
                 if capi_config.header.generation {
                     build_include_file(
                         ws,
                         header_name,
-                        &cpkg.version,
+                        emit_version_constants.then_some(&cpkg.version),
                         &root_output,
                         &cpkg.root_path,
                     )?;
@@ -1275,7 +1307,7 @@ pub fn ctest(
 ) -> CliResult {
     compile_opts.build_config.requested_profile =
         args.get_profile_name("test", ProfileChecking::Custom)?;
-    compile_opts.build_config.mode = CompileMode::Test;
+    compile_opts.build_config.intent = UserIntent::Test;
 
     compile_opts.filter = ops::CompileFilter::new(
         LibRule::Default,   // compile the library, so the unit tests can be run filtered
@@ -1409,11 +1441,11 @@ mod tests {
         );
         assert_eq!(
             static_libraries(libs_msvc, &target_msvc),
-            "-lkernel32 -ladvapi32 -lntdll -luserenv -lws2_32 -lmsvcrt"
+            "-lkernel32 -ladvapi32 -lkernel32 -lntdll -luserenv -lws2_32 -lkernel32 -lws2_32 -lkernel32 -lmsvcrt"
         );
         assert_eq!(
             static_libraries(libs_mingw, &target_mingw),
-            "-lkernel32 -ladvapi32 -lntdll -luserenv -lws2_32"
+            "-lkernel32 -ladvapi32 -lkernel32 -lntdll -luserenv -lws2_32 -lkernel32 -lws2_32 -lkernel32"
         );
     }
 }
